#!/usr/bin/env python
# -*- coding: utf_8 -*-

# based on https://github.com/pymodbus-dev/pymodbus/blob/dev/examples/client_sync.py

import logging
import time
import sys
import struct
import pymodbus
from pymodbus.exceptions import ModbusException
from pymodbus.client import ModbusSerialClient
from  pymodbus.file_message import FileRecord
from uf2 import *

PORT = '/dev/ttyUSB0'
DEVICE_ADDR = 1
FIRMWARE = "build/heizung-release.uf2"
#FIRMWARE = "build/blinky-release.uf2"
ACTIVE_PARTITION_ADDRESS = 0x10007000

logger = logging.getLogger()
logger.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
logger.addHandler(log_handler)

def main():
    firmware_blocks = read_blocks(FIRMWARE)
    firmware_streams = [[firmware_blocks[0]]]
    prev = firmware_blocks[0]
    for block in firmware_blocks[1:]:
        if block.block_no == 0 or block.num_blocks != prev.num_blocks:
            firmware_streams.append([])
        firmware_streams[-1].append(block)
        prev = block
    logger.info("found %s streams in firmware file %s" % (len(firmware_streams), FIRMWARE))

    firmware_streams2 = []
    for i, stream in enumerate(firmware_streams):
        valid = all(block.valid for block in stream)
        relevant = any(block.target_addr >= ACTIVE_PARTITION_ADDRESS for block in stream)
        if not valid:
            raise Exception("stream %s contains invalid blocks" % i)
        if relevant:
            firmware_streams2.append(stream)
        else:
            logger.info("skipping stream %s because it contains only irrelevant blocks (e.g. bootloader)" % i)
    if len(firmware_streams2) == 0:
        raise Exception("no relevant streams")

    try:
        client = ModbusSerialClient(
            port=PORT,
            timeout=0.2,
            # Common optional paramers:
            #    framer=ModbusRtuFramer,
            #    timeout=10,
            #    retries=3,
            #    retry_on_empty=False,
            #    close_comm_on_error=False,
            #    strict=True,
            #baudrate=19200,
            baudrate=115200,
            parity="E",
        )
        client.connect()
        logger.info("connected")

        blocks_progress = 0
        blocks_total = sum(len(stream) for stream in firmware_streams2) * 3
        for stream in firmware_streams2:
            #bytes = b"".join(block.bytes for block in stream)
            for block in stream:
                maxlen = (0xf5-7) // 2 * 2
                offset = 0
                for j in range(3):
                    x = client.write_file_record([
                        FileRecord(file_number=1, record_number=offset//2, record_data=block.bytes[offset:offset+maxlen]),
                    ], slave=DEVICE_ADDR)
                    blocks_progress += 1
                    logger.info("%3s / %3s, %08x, %s", blocks_progress, blocks_total, block.target_addr, x)
                    offset += maxlen

            x = client.read_input_registers(25, 4, slave=DEVICE_ADDR)
            logger.info(x)
            logger.info(x.registers)
            if x.registers != [0, 1, 0, 1]:
                raise Exception("We haven't implemented re-sending blocks, yet.")

        if False:
            with open("flash-read.bin", "wb") as f:
                block_size = 32768
                for offset in range(0, 2*1024*1024, block_size):
                    x = client.write_registers(3, offset//512, slave=DEVICE_ADDR)
                    logger.info("set offset: %r" % x)

                    read_size = 200
                    for offset2 in range(0, block_size, read_size):
                        r = FileRecord(file_number=1, record_number=offset2//2, record_length=min(read_size, block_size-offset2)//2)
                        x = client.read_file_record([
                            r
                        ], slave=DEVICE_ADDR)
                        #logger.info("read: %r" % x)
                        #logger.info("read: %r, %r" % (x.records[0], r))
                        #logger.info("read: %r" % x.records[0].record_data)
                        data = x.records[0].record_data
                        assert(len(data) == r.record_length*2)
                        assert(f.tell() == offset+offset2)
                        f.write(data)

        x = client.write_registers(1, struct.unpack(">H", b"UP")[0], slave=DEVICE_ADDR)
        logger.info("write to mark for update: %s", x)

        x = client.write_registers(1, struct.unpack(">H", b"RE")[0], slave=DEVICE_ADDR)
        logger.info("write to reset: %s", x)

    except ModbusException as exc:
        logger.error("%s", exc)

if __name__ == "__main__":
    main()
